--  font.e
--  jiri_babor@hotmail.com
--  00-09-02
--  version 6.00

without type_check
without warning

-- include image.e
include eu_engin.e


global constant ROM = 1,    -- 8x16 rom font - default

--  attributes are 'additive', e.g.: BOLD+SHADOW=1+4=5
    NONE        = 0,        -- plain
    BOLD        = 1,
    ITALIC      = 2,
    SHADOW      = 4,
    UNDERLINE   = 8,
    HILITE      = 16,

--  (multi-)line justification
    LEFT        = 0,
    CENTRE      = 1,
    RIGHT       = 2,
    BOXED       = 4

object
    fc,             -- text foreground color or 2d pattern (tile)
    hc              -- hilite color

global sequence all_fonts
sequence
    font,           -- currently loaded font image
    io,             -- italic line offset
    width           -- char widths of current font (in pixels)

integer
    at,             -- current combined attributes
    baseline,       -- baseline height of current font
    byte_width,     -- max byte width of current font
    bold,           -- boldface flag (also is bold offset)
    ecp,            -- extra char pitch (extra kerning space)
    elp,            -- extra line pitch (extra line spacing)
    f,              -- current font handle
    Height,         -- nominal height of current font
    hilite,         -- hilite flag
    hso,            -- horizontal shadow offset
    italic,         -- italic flag
    ju,             -- text justification (see global constants above)
    pc,             -- paper color, prompt background
    sc,             -- shadow color
    shadow,         -- shadow flag
    underline,      -- underline flag
    vso,            -- vertical shadow offset
    x,              -- text pointer: left edge char space
    y               -- text pointer: top  edge char space


global function rtrim(sequence s)
--  trim (discard) trailing zeros of sequence s

    integer i

    i=length(s)
    while i do
	if s[i] then exit end if
	i-=1
    end while
    return s[1..i]
end function

function expand_pattern(object c)
    integer maxwidth

    if sequence(c) then
	maxwidth = 8*byte_width + floor(Height/4) + ecp + 1
	for i=1 to length(c) do
	    while length(c[i]) < maxwidth do c[i] &= c[i] end while
	end for
	while length(c) < Height do c &= c end while
    end if
    return c
end function -- expand_pattern

function rle(sequence s)
--  return run-length-encoded *bit* sequence s
--  trailing zeros are not counted, simply trimmed off
--  count always starts with count of leading zeros

    sequence u
    integer flag,sum

    s=rtrim(s)
    flag=1
    sum=0
    u={}
    for i=1 to length(s) do
	if flag xor s[i] then
	    sum+=1
	else
	    u&=sum
	    flag=not flag
	    sum=1
	end if
    end for
    if sum then u&=sum end if
    return u
end function -- rle

function is_loaded(sequence name)
--  returns allocated handle if font is already loaded, else zero

    for i=1 to length(all_fonts) do
	if equal(name,all_fonts[i][1]) then
	    return i
	end if
    end for
    return 0
end function -- is_loaded

global procedure align(integer previous_font)
--  align newly selected font with previous_font
    y += all_fonts[previous_font][5] - baseline
end procedure -- align

global function height()
    return Height
end function

global function base_height()
    return baseline
end function

global function attributes()
    return at
end function

global procedure set_attributes(integer a)
    at = a
    bold = and_bits(at,BOLD)                -- new bold flag & offset
    italic = and_bits(at,ITALIC) != 0       -- new italic flag
    shadow = and_bits(at,SHADOW) != 0       -- new shadow flag
    underline = and_bits(at,UNDERLINE) != 0 -- new underline flag
    hilite = and_bits(at,HILITE) != 0       -- new hilite flag
end procedure

global procedure select_font(integer fh)
--  select font with handle fh, make it current

    if fh > 0 and fh <= length(all_fonts) and f != fh then
	f = fh                          -- current handle
	font = all_fonts[f][2]          -- font images
	width = all_fonts[f][3]         -- widths of current font
	Height = all_fonts[f][4]        -- height of font
	baseline = all_fonts[f][5]      -- baseline height of font
	byte_width = all_fonts[f][6]

	-- calculate italic offsets
	io = repeat(0, Height)
	for i=1 to Height do
	    io[i] = floor((Height-i)/4)
	end for
	fc = expand_pattern(fc)
	hc = expand_pattern(hc)
    end if
end procedure -- select_font

global procedure reset_font()
--  reset entire environment of current font to defaults
    x = 0               -- left edge of char space
    y = 0               -- top edge of char space
    at = NONE           -- font attributes, none
    bold = 0            -- flag
    italic = 0          -- flag
    shadow = 0          -- flag
    hilite = 0          -- flag
    underline = 0       -- flag
    fc = BRIGHT_WHITE   -- foreground text color (bright white)
    sc = GRAY           -- shadow color (gray)
    hc = GREEN          -- hilite color
    ju = LEFT           -- default justification
    ecp = 0             -- extra char pitch
    elp = 0             -- extra line pitch
    hso = 1             -- horizontal shadow offset, right
    vso = 1             -- vertical shadow offset, down
end procedure -- reset_font

global function load_font(sequence filename)
--  load font file into memory and make it available for selection

    integer bh,bw,fc,first_char,fn,he,junk,last_char,n,size
    sequence fo,row_bits,cs,cbs,wi

    n = is_loaded(filename)
    if n then
	return n
    else
	fn = open(filename, "rb")
	if fn = -1 then
	    junk = graphics_mode(-1)
	    puts(1, "Font load error: " & filename & " not found !\n")
	    abort(1)
	end if
	fc = getc(fn)
	last_char = getc(fn)
	bw = getc(fn)
	he = getc(fn)               -- nominal height
	junk = getc(fn)             -- version number, unused
	bh = getc(fn)               -- baseline height
	if bh=0 then bh=he end if   -- emergency solution for pre 3.xx fonts

	-- get widths
	wi=repeat(0,255)
	if fc = 0 then              -- skip zero char
	    first_char = 1
	    junk = getc(fn)         -- discard zero char
	else first_char = fc
	end if
	for i=first_char to last_char do
	    wi[i] = getc(fn)
	end for

	-- get char images
	fo = repeat({},255)
	size = he*bw
	if fc=0 then                -- again skip zero char
	    for i=1 to size do      -- dump
		junk = getc(fn)
	    end for
	end if
	for i = first_char to last_char do
	    cbs = repeat(0,size)    -- char byte sequence
	    cs = {}                 -- char sequence of row bit sequences
	    for j=1 to size do
		cbs[j] = getc(fn)
	    end for
	    cbs = rtrim(cbs)        -- remove trailing (bottom) bytes
	    row_bits = {}
	    for j=1 to length(cbs) do
		row_bits &= (and_bits(cbs[j],{128,64,32,16,8,4,2,1}) and 1)
		if remainder(j,bw)=0 then
		    cs=append(cs,rle(rtrim(row_bits)))
		    row_bits={}
		end if
	    end for
	    if length(row_bits) then
		cs=append(cs,rle(rtrim(row_bits)))
	    end if
	    fo[i] = cs
	end for
	close(fn)
	all_fonts = append(all_fonts,{filename,fo,wi,he,bh,bw})
	return length(all_fonts)
    end if
end function  -- load_font

procedure load_rom_font()
    sequence r, cb
    atom a

--  first load : initialize fonts
--  load 8x16 font
    all_fonts = {}
    f = load_font("vga8x16.e")

end procedure -- load_rom_font

global function lengths(sequence text)
--  return sequence of lengths, in pixels, of substrings
--  of text (using current font, attributes and parameters)
--  substrings are delimited by embedded newline chars
--  extra lengths of shadows are *ignored*

    object len
    integer a,b,c,l,w,it,hf

    len = {}   l = 0
    b = bold                            -- current bold flag/offset
    a = at                              -- current combined attributes
    it = italic                         -- current italic flag
    hf = hilite                         -- current hilite flag
    for i=1 to length(text) do
	c=text[i]
	if c<=0 then
	    a = -c
	    b = and_bits(a,BOLD)        -- new bold offset
	    it = and_bits(a,ITALIC)!=0  -- new italic flag
	    hf = and_bits(a,HILITE)!=0  -- new hilite flag
	elsif c = 10 then               -- new line
	    if l then
		if it then
		    l += io[1]          -- extra italic length of last char
		end if
	    end if
	    len &= l                    -- append length of previous line
	    l=0                         -- starting length of new line
	elsif c = '~' and hf then       -- do nothing
	elsif c<256 then                -- writable char
	    w=width[c]
	    if w then l += w+ecp+b end if
	end if
    end for
    if l then
	if it then
	    l += io[1]                  -- extra italic length of last char
	end if
    end if
    return len & l                      -- append also length of last line
end function   -- lengths

function substrings(sequence s)
--  return sequence of substrings
--  substrings are separated by embedded newlines
    sequence os
    integer i,l
    os={}
    l=length(s)
    if l then
	i=1
	for j=1 to l do
	    if s[j]=10 then
		os=append(os,s[i..j])
		i=j+1
	    end if
	end for
	if i<=l then os=append(os,s[i..l]) end if
    end if
    return os
end function -- substrings

global procedure write_char(integer ch)
--  print single character using current font attributes

    sequence c,cr
    integer u,v,w

    w=width[ch]
    if w then
	c=font[ch]          -- rle char image
	if shadow then
	    v=y+vso
	    for r=1 to length(c) do
		u=x+hso
		if italic then
		    u+=io[r]
		end if
		cr=c[r]         -- row of char image
		for j=1 to length(cr) by 2 do
		    u+=cr[j]
		    pixel(repeat(sc,cr[j+1]+bold),{u,v})
		    u+=cr[j+1]
		end for
		v+=1
	    end for
	    if underline then
		pixel(repeat(sc,w+bold+ecp),{x+hso, y+Height+vso-1})
	    end if
	end if
	v=y
	for r=1 to length(c) do
	    u=x
	    if italic then
		u+=io[r]
	    end if
	    cr=c[r]         -- row of char image
	    for j=1 to length(cr) by 2 do
		u+=cr[j]
		if atom(fc) then
		    pixel(repeat(fc,cr[j+1]+bold),{u,v})
		else
		    pixel(fc[r][u-x+1..u-x+cr[j+1]+bold],{u,v})
		end if
		u+=cr[j+1]
	    end for
	    v+=1
	end for
	if underline then
	    if atom(fc) then
		pixel(repeat(fc,w+bold+ecp),{x,y+Height-1})
	    else
		pixel(fc[Height][1..w+bold+ecp],{x,y+Height-1})
	    end if
	end if
	x+=w+bold+ecp       -- index to next char
    end if
end procedure -- char

global procedure write(sequence s)
    object o
    sequence lens,string,strings
    integer ch,len,xo

    strings = substrings(s)
    lens = lengths(s)
    if ju>2 then                    -- BOXED
	y -= floor(length(lens)*(Height+elp)/2)
    end if
    xo = x
    for i=1 to length(strings) do
	string = strings[i]
	len=lens[i]
	if and_bits(ju, RIGHT) then
	    x=xo-len
	elsif and_bits(ju, CENTRE) then
	    x=floor(xo-(len+1)/2)
	else x=xo                    -- LEFT
	end if
	for j=1 to length(string) do
	    ch=string[j]
	    if ch<=0 then set_attributes(-ch)
	    elsif ch>255 then fc=ch-256     -- change foreground color
	    elsif ch = 10 then              -- newline
		y += Height + elp
		x = xo
	    elsif ch = '~' and hilite then  -- swap hilite and foreground
		o = fc
		fc = hc
		hc = o
	    else                    -- writable char
		write_char(ch)
	    end if
	end for
    end for
end procedure   -- write

global procedure setx(integer i)
    x=i
end procedure

global procedure sety(integer i)
    y=i
end procedure

global procedure setxy(integer newx,integer newy)
    x=newx y=newy
end procedure

global function getx()
    return x
end function

global function gety()
    return y
end function

global function char_pitch(integer c)
    return width[c]+ecp+bold
end function

global function line_pitch()
    return Height + elp
end function

global function foreground()
    return fc
end function

global procedure set_foreground(object c)
    fc = expand_pattern(c)
end procedure -- set_foreground

global function shadow_color()
    return sc
end function

global procedure set_shadow_color(integer c)
    sc = c
end procedure

global function hilite_color()
    return hc
end function

global procedure set_hilite_color(object c)
    hc = expand_pattern(c)
end procedure

global procedure set_paper_color(integer c)
    pc = c
end procedure

global function paper_color()
    return pc
end function

global procedure set_shadow_offsets(integer xo, integer yo)
    hso = xo
    vso = yo
end procedure

global function shadow_offsets()
    return {hso, vso}
end function

global function justification()
    return ju
end function

global procedure justify(integer j)
    ju=j
end procedure

global procedure set_extra_line_pitch(integer p)
    elp = p
end procedure

global procedure set_extra_char_pitch(integer p)
    ecp=p
end procedure

procedure xor_cursor(integer x, integer y, integer h)
-- vertical line 2 pixels wide

    for i=0 to h-1 do
	pixel(xor_bits(get_pixel({x,y+i,2}),15),{x,y+i})
    end for
end procedure

function Length(sequence s)
-- *primitive* string length in pixels

    integer len
    len=0
    for i=1 to length(s) do
	len += width[s[i]]
    end for
    return len
end function

procedure Write(sequence s)
--  *primitive* write

    integer c

    for i=1 to length(s) do
	c=s[i]
	if c>0 and c<256 then
	    write_char(c)
	end if
    end for
end procedure

global function prompt(
    sequence s,             -- initial prompt string, editable
    integer w               -- maximum field width, in pixels
    )
    atom t
    integer
	fci,                -- first char index
	flag,               -- cursor flag, drawn/erased
	lastchar,
	lcw,                -- last raw_char width
	len,                -- current length of string
	key,                -- pressed key
	sat,                -- current attributes
	sju,                -- current line justification
	sx                  -- starting x, left edge of box

    sx = x                  -- save initial x coordinate
    sat = at                -- save initial attributes
    set_attributes(NONE)
    sju = ju                -- save initial justification
    ju = LEFT

    -- generate background strip : polygon
    polygon(pc,1, {{x,y},{x+w-1,y},{x+w-1,y+Height-1},{x,y+Height-1}})

    fci=1
    len=length(s)
    while Length(s[fci..len])>w do
	fci+=1
    end while
    write(s[fci..len])

    while 1 do
	-- blinking xor cursor
	t=time()+0.1
	flag=0                      -- cursor not drawn
	while 1 do
	    if time()-t > 0.1 then
		t=time()
		xor_cursor(x,y,Height)
		flag=not flag
	    end if
	    key=get_key()
	    if key!=-1 then
		if flag then        -- erase cursor if shown
		    xor_cursor(x,y,Height)
		end if
		exit
	    end if
	end while

	if find(key,{9,13,27,271,328,336}) then exit
	elsif key=8 then            -- backspace
	    if len then
		lastchar=s[len]
		lcw=width[lastchar]
		len-=1
		s=s[1..len]
		while fci>1 and Length(s[fci..len])+width[s[fci-1]]<=w do
		    fci-=1
		end while

		-- erase
		polygon(pc,1,{{sx,y},{sx+w-1,y},{sx+w-1,y+Height-1},{sx,y+Height-1}})
		x=sx
		write(s[fci..len])      -- reprint visible tail
	    end if
	elsif key>0 and key<256 then
	    s &= key
	    len+=1
	    if Length(s[fci..len])>w then
		while Length(s[fci..len])>w do
		    fci+=1
		end while

		-- erase
		polygon(pc,1,{{sx,y},{sx+w-1,y},{sx+w-1,y+Height-1},{sx,y+Height-1}})
		x=sx
		write(s[fci..len])      -- print visible tail
	    else write_char(key)
	    end if
	end if
    end while
    set_attributes(sat)                 -- restore initial attributes
    justify(sju)                        -- restore initial justification
    return {key,s}
end function   -- prompt

global procedure write_within(sequence s, integer n, integer len)
--  write string s starting with its nth char using current font setting
--  including changes in first n-1 characters
--  truncate string if total length in pixels exceeds len

    sequence l,t
    integer c

    if n > length(s) then n = length(s) end if  -- for safety
    if n < 1 then n = 1 end if
    if n > 1 then
	for i=1 to n-1 do
	    c = s[i]
	    if c <= 0 then
		set_attributes(-c)
	    end if
	    if c > 255 then
		set_foreground(c-256)
	    end if
	end for
    end if
    t = s[n..length(s)]
    l = lengths(t)
    while l[1] > len do
	t = t[1..length(t)-1]
	l = lengths(t)
    end while
    write(t)
end procedure -- write_within


-- initialize font : defaults ----------------------------------------
load_rom_font()
reset_font()

